﻿//Grant J. Combs
//Senior Project 2021
//Allows a user to create a cryptogram using a variety of ciphers. 

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;
using System.Threading.Tasks;
using System.Windows.Forms;
using System.IO;
using WindowsFormsApp1.Properties;
using ExtensionMethods;
using System.Diagnostics;

namespace WindowsFormsApp1
{
    public partial class cipherEncoder : Form
    {
        bool unsavedChanges = false;
        string fn = "";
        Regex alpha = new Regex("[a-zA-Z]");
        Regex uppercase = new Regex("[A-Z]");
        Regex lowercase = new Regex("[a-z]");

        string file = Resources.google_10000_english_usa;
        string[] englishWords = Resources.google_10000_english_usa.ParseWords();

        public cipherEncoder()
        {
            InitializeComponent();
            DisableOptions();
            outputBox.ReadOnly = true;
            debugWindow.ReadOnly = true;
            scoreDisplay.Visible = false;
            Console.SetOut(new ControlWriter(debugWindow));
        }

        //Checks for changes, prompts the user if there are any, and then exits the program.
        private void ExitToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (unsavedChanges)
            {
                DialogResult exitPrompt = MessageBox.Show("The output window contains unsaved text. Would you like to save your output?", "Unsaved Output", MessageBoxButtons.YesNo, MessageBoxIcon.Question);
                if (exitPrompt == DialogResult.Yes)
                    Save(sender, e);
            }
            else if (inputBox.Text != "")
            {
                DialogResult exitPrompt = MessageBox.Show("The input window currently contains text. Are you sure you want to exit?", "Non-Empty Input", MessageBoxButtons.YesNo, MessageBoxIcon.Question);
                if (exitPrompt == DialogResult.Yes)
                    Application.Exit();
                else Save(sender, e);
            }
            Application.Exit();
        }

        //Allows copy to be used in the GUI
        private void CopyToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (this.ActiveControl is TextBox)
            {
                TextBox myBox = (TextBox)this.ActiveControl;
                Clipboard.SetDataObject(myBox.SelectedText, true);
            }
        }

        //Allows paste to be used in the GUI
        private void Paste_Click(object sender, EventArgs e)
        {
            TextBox myBox = (TextBox)this.ActiveControl;
            int start = myBox.SelectionStart;
            myBox.Text = myBox.Text.Insert(start, Clipboard.GetText());
        }

        //Allows cut to be used in the GUI
        private void Cut_Click(object sender, EventArgs e)
        {
            if (this.ActiveControl is TextBox)
            {
                TextBox myBox = (TextBox)this.ActiveControl;
                int length = myBox.SelectionLength;
                Clipboard.SetDataObject(myBox.SelectedText, true);
                myBox.Text = myBox.Text.Remove(myBox.SelectionStart, length);
            }
        }

        //Handles the removing of spaces, symbols, or both.
        private void RemoveThings(object sender, EventArgs e)
        {
            ToolStripMenuItem tmp = sender as ToolStripMenuItem;
            string s;
            if (tmp.Name == "removeSpacesAndSymbols") s = "[^0-9a-zA-Z]";
            else s = (tmp.Name == "removeSpaces") ? " " : "[^0-9a-zA-Z ]";
            inputBox.Text = Regex.Replace(inputBox.Text, s, "");
        }

        //Opens a file to be used as input
        private void OpenFile(object sender, EventArgs e)
        {
            OpenFileDialog openFileWindow = new OpenFileDialog();
            openFileWindow.Title = "Open input file";
            openFileWindow.Filter = "Text files (*.txt)|*.txt|All files (*.*)|*.*";
            if (openFileWindow.ShowDialog() == DialogResult.OK)
            {
                try
                {
                    if (inputBox.Text != "") inputBox.Clear();
                    string[] fileLines = File.ReadAllLines(openFileWindow.FileName);
                    for (int i = 0; i < fileLines.Length; i++)
                        inputBox.Text += (fileLines[i] + "\r\n");
                }
                catch (IOException)
                {
                    MessageBox.Show("I/O error. Please try again.", "I/O Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                }
            }
        }

        //Converts input to either all lowercase or all uppercase
        private void Convert(object sender, EventArgs e)
        {
            ToolStripMenuItem tmp = sender as ToolStripMenuItem;
            string[] lines = inputBox.Lines;
            for (int i = 0; i < inputBox.Lines.Length; i++)
                lines[i] = (tmp.Name == "convertAllToUppercase") ? inputBox.Lines[i].ToUpper() : inputBox.Lines[i].ToLower();
            inputBox.Lines = lines;
        }

        //Handles the "Save" button- will save to the previously written to file or, if there is none, will prompt the user to create an output file
        private async void Save(object sender, EventArgs e)
        {
            if (fn == "") SaveAs(sender, e);
            else
            {
                try
                {
                    await Task.Run(() => File.WriteAllLines(fn, outputBox.Lines)); //this is written asynchronously so saving will not hang up the program (mostly for big files)
                    MessageBox.Show("File successfully saved.", "File saved", MessageBoxButtons.OK, MessageBoxIcon.Information);
                    unsavedChanges = false;
                }
                catch (IOException)
                {
                    MessageBox.Show("I/O error. Please try again.", "I/O Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                }
            }

        }

        //Handles the "Save As" button- will always prompt the user to write a new output file
        private async void SaveAs(object sender, EventArgs e)
        {
            if (outputBox.Text == "")
                MessageBox.Show("The Output window is currently empty. Please encode a message first.", "Error: Empty Output Box", MessageBoxButtons.OK, MessageBoxIcon.Error);
            else
            {
                SaveFileDialog saveFileWindow = new SaveFileDialog();
                saveFileWindow.Title = "Save output as...";
                saveFileWindow.Filter = "Text files (*.txt)|*.txt|All files (*.*)|*.*";
                try
                {
                    if (saveFileWindow.ShowDialog() == DialogResult.OK)
                    {
                        fn = saveFileWindow.FileName;
                        await Task.Run(() => File.WriteAllLines(fn, outputBox.Lines));
                        MessageBox.Show("File successfully saved.", "File saved", MessageBoxButtons.OK, MessageBoxIcon.Information);
                        unsavedChanges = false;
                    }
                }
                catch (IOException)
                {
                    MessageBox.Show("I/O error. Please try again.", "I/O Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                }
            }
        }

        //Clears the input or output - specifically for the "Clear" buttons
        private void clear(object sender, EventArgs e)
        {
            if ((sender as Button).Name == "clearInput") inputBox.Clear();
            else outputBox.Clear();
        }

        //Changes the visible options depending on which cipher is chosen
        private void EncodeMethodChanged (object sender, EventArgs e)
        {
            switch (method.SelectedItem.ToString())
            {
                case "Caesar/Additive (monoalphabetic)":
                    DisableOptions();
                    shiftBy.Enabled = true;
                    shiftByBox.Enabled = true;
                    break;
                case "Multiplicative (monoalphabetic)":
                    DisableOptions();
                    shiftFactorByBox.Enabled = true;
                    shiftFactorBy.Enabled = true;
                    break;
                case "Affine (monoalphabetic)":
                    DisableOptions();
                    shiftFactorByBox.Enabled = true;
                    shiftFactorBy.Enabled = true;
                    slopeBox.Enabled = true;
                    slopeLabel.Enabled = true;
                    shiftFactorBy.Text = "m = ";
                    warningMessage.Text = "Please enter your parameters in slope-intercept form (y  = mx + b).";
                    warningMessage.Visible = true;
                    break;
                case "Hill Digraph (polyalphabetic)":
                    DisableOptions();
                    keywordLabel.Enabled = true;
                    keywordBox.Enabled = true;
                    warningMessage.Text = "Keyword must be four letters. If < 4, letters will be added.";
                    warningMessage.Visible = true;
                    break;
                case "Vigenere (polyalphabetic)":
                case "Playfair (polyalphabetic)":
                    DisableOptions();
                    keywordLabel.Enabled = true;
                    keywordBox.Enabled = true;
                    break;
                default:
                    DisableOptions();
                    break;
            }
        }

        //Hides all additional options besides the cipher dropdown menu.
        private void DisableOptions()
        {
            keywordLabel.Enabled = false;
            keywordBox.Enabled = false;
            shiftBy.Enabled = false;
            shiftByBox.Enabled = false;
            shiftFactorByBox.Enabled = false;
            shiftFactorBy.Enabled = false;
            slopeBox.Enabled = false;
            slopeLabel.Enabled = false;
            warningMessage.Visible = false;
            shiftFactorBy.Text = "Shift by a factor of: ";
        }

        //Handles the "Encode" and "Decode" button- depending on the encoding options will call different methods.
        private void doWork(object sender, EventArgs e)
        {
            if (inputBox.Text == "")
            {
                MessageBox.Show("Please enter an input first.", "Error: No Input", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
                return;
            }
            else if (outputBox.Text.Length != 0)
            {
                DialogResult dialogResult = MessageBox.Show("The output box is not empty. Would you like to erase its contents and continue?", "Overwrite previous output?", MessageBoxButtons.YesNo, MessageBoxIcon.Question);
                if (dialogResult == DialogResult.No) return;

            }
            if (method.GetItemText(method.SelectedItem) == "")
            {
                MessageBox.Show("Please select a cipher to use.", "Error: no cipher selected", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
                return;
            }
            scoreDisplay.Visible = false;
            string[] output = new string[inputBox.Lines.Length];
            string keyword = keywordBox.Text;
            if (keyword != "")
            {
                Remove(ref keyword, "all");
                keywordBox.Text = keyword;
            }
            List<object> errors = new List<object>();
            bool mode = ((sender as Button).Name == "encode") ? false : true;
            string msg = "Starting ";
            msg += (mode == false) ? "Encryption " : "Decryption ";
            msg += "with " + method.GetItemText(method.SelectedItem);
            Console.WriteLine(msg);
            Stopwatch watch = Stopwatch.StartNew();
            switch (method.GetItemText(method.SelectedItem))
            {
                case "Caesar/Additive (monoalphabetic)":
                    errors = OptionRequirementsMet(shiftByBox);
                    if (errors.Count == 0) Caesar_Cipher(ref output, mode, int.Parse(shiftByBox.GetItemText(shiftByBox.SelectedItem))); ;
                    break;
                case "Multiplicative (monoalphabetic)":
                    errors = OptionRequirementsMet(shiftFactorByBox);
                    if (errors.Count == 0) Multiplicative_Cipher(ref output, mode, int.Parse(shiftFactorByBox.GetItemText(shiftFactorByBox.SelectedItem)));
                    break;
                case "Affine (monoalphabetic)":
                    errors = OptionRequirementsMet(shiftFactorByBox, slopeBox);
                    if (errors.Count == 0) Affine_Cipher(ref output, mode, int.Parse(shiftFactorByBox.GetItemText(shiftFactorByBox.SelectedItem)), int.Parse(slopeBox.GetItemText(slopeBox.SelectedItem)));
                    break;
                case "Hill Digraph (polyalphabetic)":
                    errors = OptionRequirementsMet(keywordBox);
                    if (errors.Count == 0) HillDigraph_Cipher(ref output, mode, ref keyword);
                    break;
                case "Vigenere (polyalphabetic)":
                    errors = OptionRequirementsMet(keywordBox);
                    if (errors.Count == 0) Vigenere_Cipher(ref output, mode, keyword);
                    break;
                case "Playfair (polyalphabetic)":
                    errors = OptionRequirementsMet(keywordBox);
                    if (errors.Count == 0) Playfair_Cipher(ref output, mode, ref keyword);
                    break;
                default:
                    MessageBox.Show("Please select a cipher to use.", "Error: no cipher selected", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
                    break;
            }

            watch.Stop();
            Console.WriteLine("Operation completed. Time elapsed: " + watch.ElapsedMilliseconds + "ms.\r\n");
            if (errors.Count == 0)
            {
                unsavedChanges = true;
                outputBox.Lines = output;
            }
            else DisplayRequirementError(errors);
        }

        //Determines whether all the required options for encoding are satisfied. Returns a list of all the unsatisfied requirements.
        private List<object> OptionRequirementsMet(params object[] o)
        {
            List<object> errors = new List<object>();
            for (int i = 0; i < o.GetLength(0); i++)
            {
                if (o[i] is ComboBox && (o[i] as ComboBox).SelectedIndex == -1)
                    errors.Add(o[i] as ComboBox);
                else if (o[i] is TextBox && String.IsNullOrWhiteSpace((o[i] as TextBox).Text))
                    errors.Add(o[i] as TextBox);
            }
            return errors;
        }

        //Displays the error message that displays when one of the object fields is left blank.
        private void DisplayRequirementError(List<object> o)
        {
            string errorMessage = "Error: the following option(s) were not specified:\r\n\r\n";
            for (int i=0; i < o.Count; i++)
            {
                if (o[i] is ComboBox)
                {
                    switch ((o[i] as ComboBox).Name)
                    {
                        case "shiftByBox":
                            errorMessage += "•  Shift By\r\n";
                            break;
                        case "shiftFactorByBox":
                            if (method.GetItemText(method.SelectedItem) == "Affine (monoalphabetic)") errorMessage += "•  Slope (m = )\r\n";
                            else errorMessage += "•  Shift Factor By\r\n";
                            break;
                        case "slopeBox":
                            errorMessage += "•  Intercept (b =)\r\n";
                            break;
                    }
                }
                else
                    errorMessage += "•  Keyword\r\n";
            }
            MessageBox.Show(errorMessage, "Error: unspecified options", MessageBoxButtons.OK, MessageBoxIcon.Error);
        }

        //Handles the Caesar cipher. Mode boolean 0 = encryption, 1 = decryption
        private void Caesar_Cipher(ref string[] output, bool mode, params int[] num)
        {
            int multiplier = (mode == false) ? 1 : -1;
            for (int i = 0; i < inputBox.Lines.Length; i++)
                for (int j = 0; j < inputBox.Lines[i].Length; j++)
                    if (!alpha.IsMatch(inputBox.Lines[i][j].ToString())) output[i] += inputBox.Lines[i][j];
                    else output[i] += (char)(Mod((Position(inputBox.Lines[i][j]) + (multiplier * num[0])), 26) + 65);
        }

        //Returns if a number is coprime with 26.
        private void IsCoPrime(object sender, EventArgs e)
        {
            if (method.GetItemText(method.SelectedItem) == "Multiplicative (monoalphabetic)" || method.GetItemText(method.SelectedItem) == "Affine (monoalphabetic)")
            {
                int a = int.Parse(shiftFactorByBox.GetItemText(shiftFactorByBox.SelectedItem));
                warningMessage.Visible = false;
                for (int i = 2; i <= a; i++) //the middle part would normally be i <= a && i <= 26, but we know based on program design that i will always be less than 26.
                    if (a % i == 0 && 26 % i == 0)
                    {
                        if (method.GetItemText(method.SelectedItem) == "Affine (monoalphabetic)") warningMessage.Text = "Please enter your parameters in slope-intercept form (y  = mx + b).\r\nWarning: this number is not coprime with 26. It will not produce a monoalphabetic cipher.";
                        else warningMessage.Text = "Warning: this number is not coprime with 26. It will not produce a monoalphabetic cipher.";
                        warningMessage.Visible = true;
                        break;
                    }
            }
        }

        //Finds the inverse key for a ciphered text when the parameters of encryption are already known.
        private int InverseKey(int k)
        {
            for (int i = 2; i < 26; i++)
                if ((i * k) % 26 == 1) return i;
            return k;
        }

        //Handles multiplicative cipher. false = encryption, true = decryption
        private void Multiplicative_Cipher(ref string[] output, bool mode, params int[] num)
        {
            int key = (mode == false) ? num[0] : InverseKey(num[0]);
            for (int i = 0; i < inputBox.Lines.Length; i++)
                for (int j = 0; j < inputBox.Lines[i].Length; j++)
                    if (!alpha.IsMatch(inputBox.Lines[i][j].ToString())) output[i] += inputBox.Lines[i][j];
                    else output[i] += (char)(Mod((Position(inputBox.Lines[i][j]) * key), 26) + 65);
        }

        //Handles Affine cipher. false = encryption, true = decryption
        private void Affine_Cipher(ref string[] output, bool mode, params int[] num)
        {
            if (mode == false)
            {
                for (int i = 0; i < inputBox.Lines.Length; i++)
                    for (int j = 0; j < inputBox.Lines[i].Length; j++)
                        if (!alpha.IsMatch(inputBox.Lines[i][j].ToString())) output[i] += inputBox.Lines[i][j];
                        else output[i] += (char)(Mod((Position(inputBox.Lines[i][j]) * num[0] + num[1]), 26) + 65);
            }
            else
            {
                for (int i = 0; i < inputBox.Lines.Length; i++)
                    for (int j = 0; j < inputBox.Lines[i].Length; j++)
                        if (!alpha.IsMatch(inputBox.Lines[i][j].ToString())) output[i] += inputBox.Lines[i][j];
                        else output[i] += (char)(Mod((Position(inputBox.Lines[i][j]) - num[1]) * InverseKey(num[0]), 26) + 65);
            }
        }

        //Handles Vigenere Cipher, 0 = encryption, 1 = decryption
        private void Vigenere_Cipher(ref string[] output, bool mode, string key)
        {
            int multiplier = (mode == false) ? 1 : -1;
            int it = 0;
            for (int i = 0; i < inputBox.Lines.Length; i++)
                for (int j = 0; j < inputBox.Lines[i].Length; j++)
                    if (!alpha.IsMatch(inputBox.Lines[i][j].ToString())) output[i] += inputBox.Lines[i][j];
                    else
                    {
                        if (!alpha.IsMatch(inputBox.Lines[i][j].ToString())) output[i] += inputBox.Lines[i][j];
                        else
                        {
                            output[i] += (char)(Mod((Position(inputBox.Lines[i][j])) + multiplier * Position(key[Mod(it, key.Length)]), 26) + 65);
                            it++;
                        }
                    }
        }

        //Hill Digraph cipher (polyalphabetic) - this involves 2x2 matrix multiplcation. false = encryption, true = decryption
        private void HillDigraph_Cipher(ref string[] output, bool mode, ref string keyword)
        {

            if (keyword.Length < 4) keyword += "ABCD".Substring(0, 4 - keyword.Length);
            if (keyword.Length > 4) keyword = keyword.Substring(0, 4);
            keywordBox.Text = keyword;

            //Converts the keyword into a 2x2 matrix: "ABCD" becomes [A,B][C,D]
            int[,] keywordMatrix = new int[2, 2];
            int n = 0;
            for (int i = 0; i < 2; i++)
                for (int j = 0; j < 2; j++)
                {
                    if (uppercase.IsMatch(keyword[i + j].ToString())) keywordMatrix[i, j] = keyword[n] - 65;
                    else keywordMatrix[i, j] = keyword[n] - 97;
                    n++;
                }
            if (mode == true) //If we are decrypting, we need to turn this matrix into its inverse. A^-1 = 1/det(A) * adj(A)
            {
                int determinant = (keywordMatrix[0, 0] * keywordMatrix[1, 1]) - (keywordMatrix[0, 1] * keywordMatrix[1, 0]);
                int key = InverseKey(Mod(determinant,26));

                int[,] inverseMatrix = new int[2, 2];
                inverseMatrix[0, 0] = keywordMatrix[1, 1];
                inverseMatrix[0, 1] = keywordMatrix[0, 1] * -1;
                inverseMatrix[1, 0] = keywordMatrix[1, 0] * -1;
                inverseMatrix[1, 1] = keywordMatrix[0, 0];

                for (int i = 0; i < 2; i++)
                    for (int j = 0; j < 2; j++)
                        inverseMatrix[i, j] = Mod(inverseMatrix[i, j] * key, 26);

                keywordMatrix = inverseMatrix;
            }
            int[,] inputMatrix = DigraphCreation(false);

            //Multiply the matrices here.
            int[,] outputMatrix = new int[2, inputBox.TextLength / 2];
            MultiplyMatrices(keywordMatrix, inputMatrix, ref outputMatrix);
            for (int i = 0; i < outputMatrix.GetLength(1); i++)
            {
                output[0] += (char)(Mod(outputMatrix[0, i], 26) + 65);
                output[0] += (char)(Mod(outputMatrix[1, i], 26) + 65);
            }
        }

        //Creates an array of digraphs. the boolean "dl" refers to whether double letters need to be separated, which is used for the Playfair cipher.
        private int[,] DigraphCreation(bool dl)
        {
            string input = inputBox.Text;
            Remove(ref input, "all");
            if (dl == true)
            {
                input = Regex.Replace(input, "j|J", "i");
                for (int i = 0; i < input.Length - 1; i++)
                {
                    if (i % 2 == 0)
                        if (input[i] == input[i + 1]) input = input.Substring(0, i+1) + "X" + input.Substring(i+1);
                }
            }
            if (input.Length % 2 != 0) input += "X";
            inputBox.Text = input;
            int[,] inputMatrix = new int[2, input.Length / 2];
            for (int i = 0; i < input.Length / 2; i++)
            {
                inputMatrix[0, i] = (uppercase.IsMatch(input[2 * i].ToString())) ? input[2 * i] - 65 : input[2 * i] - 97;
                inputMatrix[1, i] = (uppercase.IsMatch(input[(2 * i) + 1].ToString())) ? input[(2 * i) + 1] - 65 : input[(2 * i) + 1] - 97;
            }
            return inputMatrix;
        }

        //Playfair Cipher (polyalphabetic)
        private void Playfair_Cipher(ref string[] output, bool mode, ref string key)
        {
            //Here we will construct a Polybius Square. It is a 5x5 grid and from left to right, we write out the letters of the key followed by the rest of the letters of the alphabet -- however, only one of each letter is present.
            int[,] polybiusSquare = new int[5, 5];
            for (int i = 0; i < 5; i++)
                for (int j = 0; j < 5; j++)
                    polybiusSquare[i, j] = -1;  //every space in polybius square initialized to -1
            bool[] letterUnused = new bool[26];
            for (int i = 0; i < 26; i++)
                letterUnused[i] = true;
            int x = 0, y = 0; //Coordinates of next available array space
            for (int i = 0; i < key.Length; i++)
            {
                GetNextAvailableArraySpot(polybiusSquare, ref x, ref y);
                int letter = Position(key[i]);
                if (letterUnused[letter])
                {
                    polybiusSquare[x, y] = letter;
                    letterUnused[letter] = false;
                }
            }
            for (int i = 0; i < letterUnused.GetLength(0); i++)
            {
                if (i == 9) i++; //skip "J"
                GetNextAvailableArraySpot(polybiusSquare, ref x, ref y);
                if (letterUnused[i])
                {
                    polybiusSquare[x, y] = i;
                    letterUnused[i] = false;
                }
            }
            //for  printing polybius square 
            Console.WriteLine("Polybius Square:");
            for (int i = 0; i < polybiusSquare.GetLength(0); i++)
            {
                for (int j = 0; j < polybiusSquare.GetLength(1); j++)
                    Console.Write((char)(polybiusSquare[i, j] + 65) + "\t");
                Console.WriteLine();
            }
            int[,] inputMatrix = DigraphCreation(true);

            //Rule 1 (double letters in digraphs should be separated by "X") is taken care of in the DigraphCreation(bool) method with setting the bool to true.
            int x1 = 0, y1 = 0, x2 = 0, y2 = 0;
            for (int i = 0; i < inputMatrix.GetLength(1); i++)
            {
                FindInPS(polybiusSquare, inputMatrix[0, i], inputMatrix[1, i], ref x1, ref y1, ref x2, ref y2);

                //Rule 2 - if two letters appear in same row in Polybius Square, then replace each letter by the one immediately to its right
                if (x1 == x2)
                {
                    if (mode == false)
                    {
                        inputMatrix[0, i] = polybiusSquare[x1, Mod(y1 + 1, 5)];
                        inputMatrix[1, i] = polybiusSquare[x2, Mod(y2 + 1, 5)];
                    }
                    else
                    {
                        inputMatrix[0, i] = polybiusSquare[x1, Mod(y1 - 1, 5)];
                        inputMatrix[1, i] = polybiusSquare[x2, Mod(y2 - 1, 5)];
                    }
                }

                //Rule 3 - if two letters appear in the same column in Polybius Square, then replace each letter by the one immediately below it
                else if (y1 == y2)
                {
                    if (mode == false)
                    {
                        inputMatrix[0, i] = polybiusSquare[Mod(x1 + 1, 5), y1];
                        inputMatrix[1, i] = polybiusSquare[Mod(x2 + 1, 5), y2];
                    }
                    else
                    {
                        inputMatrix[0, i] = polybiusSquare[Mod(x1 - 1, 5), y1];
                        inputMatrix[1, i] = polybiusSquare[Mod(x2 - 1, 5), y2];
                    }
                }

                //Rule 4 - if neither rules 2 nor 3 apply, form a rectangle in which the two letters are opposite corners. then replace with the letter that is in the other corner of the same row as that letter.
                else
                {
                    inputMatrix[0, i] = polybiusSquare[x1, y2];
                    inputMatrix[1, i] = polybiusSquare[x2, y1];

                }
            }
            //Now that we have modified the input matrix into our output, we can translate each digraph back into letters.
            for (int i = 0; i < inputMatrix.GetLength(1); i++)
            {
                output[0] += (char)(inputMatrix[0, i] + 65);
                output[0] += (char)(inputMatrix[1, i] + 65);
                output[0] += " ";
            }

        }

        //Finds a letter's coordinates in a Polybius Square.
        private void FindInPS(int[,] polybiusSquare, int c1, int c2, ref int x1, ref int y1, ref int x2, ref int y2)
        {
            for (int i=0; i < 5; i++)
                for (int j=0; j < 5; j++)
                    if (polybiusSquare[i,j] == c1)
                    {
                        x1 = i;
                        y1 = j;
                    }
                    else if (polybiusSquare[i,j] == c2)
                    {
                        x2 = i;
                        y2 = j;
                    }
        }

        //Finds the next available space in a 2D array. Used for the Playfair cipher.
        private void GetNextAvailableArraySpot(int[,] myArray, ref int x, ref int y)
        {
            for (int i = 0; i < myArray.GetLength(0); i++)
                for (int j = 0; j < myArray.GetLength(1); j++)
                    if (myArray[i, j] == -1)
                    {
                        x = i;
                        y = j;
                        return;
                    }
            x = -1;
            y = -1;
        }

        //Multiplies keyword matrix with the input matrices. 
        private void MultiplyMatrices (int[,] keywordMatrix, int[,] inputMatrix, ref int[,] outputMatrix)
        {
            for (int i = 0; i < inputMatrix.Length / 2; i++)
            {
                outputMatrix[0, i] = (keywordMatrix[0, 0] * inputMatrix[0, i]) + (keywordMatrix[0, 1] * inputMatrix[1, i]);
                outputMatrix[1, i] = (keywordMatrix[1, 0] * inputMatrix[0, i]) + (keywordMatrix[1, 1] * inputMatrix[1, i]);
            }
        }

        //Converts an ASCII character to a character's position in the alphabet. For example, "A" = ASCII 65, and the function will return position 0 of the alphabet.
        private int Position(char c)
        {
            if (uppercase.IsMatch(c.ToString())) return c - 65;
            return c - 97;
        }

        //Shows a simple About menu.
        private void aboutThisProgram(object sender, EventArgs e)
        {
            MessageBox.Show("Cipher Encrypter\r\nWritten by Grant Combs\r\n\r\n2021 Senior Project\r\nDepartment of Computer Science\r\nNorthern Michigan University\r\n\r\nList of most common English words compiled by worldlywisdom on Github from the Google Web Trillion Word Corpus. Use for educational purposes is permitted under the applicable licenses.", "About This Program", MessageBoxButtons.OK);
        }

        //Creating a modulus function because C#'s implementation of "%" is actually remainder and thus does not work with negative numbers. And C# will not let you override an already existing operator signature.
        private int Mod(int a, int b)
        {
            return (a % b + b) % b;
        }

        //Removes things from strings
        public void Remove(ref string s, params string[] stuff)
        {
            for (int i = 0; i < stuff.Length; i++)
            {
                switch (stuff[i])
                {
                    case "spaces":
                    case "space":
                        s = Regex.Replace(s, "[ ]", "");
                        break;
                    case "symbols":
                    case "symbol":
                        s = Regex.Replace(s, "[^a-zA-Z0-9]", "");
                        break;
                    case "numbers":
                        s = Regex.Replace(s, "[0-9]", "");
                        break;
                    case "all": //everything that's not a letter
                        s = Regex.Replace(s, "[^a-zA-Z]", "");
                        break;
                }
            }
        }

        //Handles automatic decryption of a cipher.

        List<result> results = new List<result>();
        int highestSoFar = 0, currentDisplayed = 0;
        int highestSoFar_j = 0;
        private void autoDecrypt(object sender, EventArgs e)
        {
            if (inputBox.Text == "")
            {
                MessageBox.Show("Please enter an input first.", "Error: No Input", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
                return;
            }
            else if (outputBox.Text != "")
            {
                DialogResult dialogResult = MessageBox.Show("The output box is not empty. Would you like to erase its contents and continue?", "Overwrite previous output?", MessageBoxButtons.YesNo, MessageBoxIcon.Question);
                if (dialogResult == DialogResult.No) return;

            }
            if (method.GetItemText(method.SelectedItem) == "")
            {
                MessageBox.Show("Please select a cipher to use.", "Error: no cipher selected", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
                return;
            }
            scoreDisplay.Visible = true;
            string[] output = new string[inputBox.Lines.Length];
            highestSoFar = 0;
            highestSoFar_j = 0;
            results.Clear();
            Console.WriteLine("Starting AutoDecrypt with " + method.GetItemText(method.SelectedItem));
            Stopwatch watch = Stopwatch.StartNew();
            switch (method.GetItemText(method.SelectedItem))
            {
                case "Caesar/Additive (monoalphabetic)":
                    for (int i = 0; i <= 25; i++)
                    {
                        output = new string[inputBox.Lines.Length];
                        Caesar_Cipher(ref output, true, i);
                        result r = new result(output, AssignScore(output, results, i, 0, false), i, 0);
                        Console.WriteLine("i= " + i + ":\tScore = " + r.score);
                        results.Add(r);
                    }
                    shiftByBox.SelectedItem = results[highestSoFar].i;
                    break;
                case "Multiplicative (monoalphabetic)":
                    for (int i = 0; i <= 25; i++)
                    {
                        output = new string[inputBox.Lines.Length];
                        Multiplicative_Cipher(ref output, true, i);
                        result r = new result(output, AssignScore(output, results, i, 0, false), i, 0);
                        Console.WriteLine("i= " + i + ":\tScore = " + r.score);
                        results.Add(r);
                    }
                    break;
                /*case "Affine (monoalphabetic)":
                    for (int i = 1; i <= 25; i++)
                        for (int j = 1; j <= 25; j++)
                        {
                            output = new string[inputBox.Lines.Length];
                            Affine_Cipher(ref output, true, i, j);
                            result r = new result(output, AssignScore(output, results, i, j, true), i, j);
                            Console.WriteLine("i= " + i + ", j = " + j + ":\tScore = " + r.score);
                            results.Add(r);
                        }

                    break;*/
                default:
                    MessageBox.Show("AutoDecrypt is not supported for this cipher.", "AutoDecrypt not supported", MessageBoxButtons.OK, MessageBoxIcon.Error);
                    Console.WriteLine("Error: AutoDecrypt does not support " + method.GetItemText(method.SelectedItem) + ". Operation not completed.");
                    return;
            }
            watch.Stop();
            Console.WriteLine("The highest result was at i=" + highestSoFar + " with score " + results[highestSoFar].score + ".");
            Console.WriteLine("Operation completed. Time elapsed: " + watch.ElapsedMilliseconds + " ms.\r\n");
            outputBox.Lines = results[highestSoFar].output;
            currentDisplayed = highestSoFar;
            scoreDisplay.Text = "Score: " + results[highestSoFar].score.ToString();
            scoreDetails.Text = "Shift: " + results[highestSoFar].i;
            if (slopeBox.Enabled == true) scoreDetails.Text += ("\r\nb = " + results[highestSoFar].j);
            unsavedChanges = true;
        }

        //Vars that are only relevant to the below method but are defined here to improve runtime
        const int DICTIONARY_SIZE = 100; //set the dictionary size for AssignScore
        string r = "", oneliner = "";
        int score = 0, l = 0; //defining these outside the function to speed up runtime

        //Assigns a score to a potential output based on how many English words are found. If it is the highest match so far, the score is assigned to highestSoFar.
        private int AssignScore(string[] output, List<result> results, int iterator, int iterator2, bool is2D)
        {
            oneliner = "";
            for (int i = 0; i < output.GetLength(0); i++)
                oneliner += output[i];
            score = 0;

            for (int i = 0; i < DICTIONARY_SIZE; i++)
            {
                r = "(?i)" + englishWords[i] + "(?-i)";
                l = englishWords[i].Length;
                if (l > 2) score += (Regex.Matches(oneliner, r).Count * l);
                else score += l;
            }

            if (is2D)
            {
                if ((iterator == 0 && iterator2 == 0) || score > results[highestSoFar].score)
                {
                    highestSoFar = iterator;
                    highestSoFar_j = iterator2;
                    Console.WriteLine("The highest score is now is at i=" + highestSoFar + ", j=" + highestSoFar_j);
                }
            }
            else if (iterator == 0 || score > results[highestSoFar].score)
            {
                highestSoFar = iterator;
                Console.WriteLine("The highest score is now is at i=" + highestSoFar);
            }

            return score;
        }

        //Allows you to iterate back and forth through generated cipher solutions.
        private void arrows(object sender, EventArgs e)
        {
            int multiplier = ((sender as Button).Name == "forward") ? 1 : -1;
            currentDisplayed = Mod(currentDisplayed + (multiplier * 1), results.Count);
            outputBox.Lines = results[currentDisplayed].output;
            scoreDetails.Text = "Shift: " + results[currentDisplayed].i;
            if (slopeBox.Enabled == true) scoreDetails.Text += ( "\r\nb = " + results[currentDisplayed].j);
            scoreDisplay.Text = "Score: " + results[currentDisplayed].score;
        }
    }

    //Custom result class for AutoDecrypt.
    public class result
    {
        public int i = 0, j = 0, score = 0;
        public string[] output;

        public result(string[] _output, int _score, int _i, int _j)
        {
            output = _output;
            score = _score;
            i = _i;
            j = _j;
        }
    }

    //This overwrites the TextWriter so that Console.WriteLine() will output to the debug window on the program interface.
    public class ControlWriter : TextWriter
    {
        private TextBox debugWindow;
        public ControlWriter(TextBox debugWindow)
        {
            this.debugWindow = debugWindow;
        }
        public override void Write(char c)
        {
            debugWindow.AppendText(c.ToString());
        }
        public override void Write(string s)
        {
            debugWindow.AppendText(s);
        }
        public override Encoding Encoding
        {
            get { return Encoding.ASCII; }
        }
    }

}

namespace ExtensionMethods
{
    public static class StringExtensions
    {
        //This takes the string of words from the resource text file and turns it into a string array for easy access.
        public static string[] ParseWords(this string file)
        {
            return file.Split(new[] { "\n" }, StringSplitOptions.None);
        }
    }
}